﻿//---------------------------------------------------------------------
// <copyright file="CommonWritingValidationScenarioTests.cs" company="Microsoft">
//      Copyright (C) Microsoft Corporation. All rights reserved. See License.txt in the project root for license information.
// </copyright>
//---------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.IO;
using Microsoft.OData.Edm;
using Xunit;

namespace Microsoft.OData.Tests.ScenarioTests.Writer
{
    public class CommonWritingValidationScenarioTests
    {
        [Fact]
        public void WriteEntryWithFakeTypeShouldFailWithUnrecognizedTypeName()
        {
            IEdmEntitySet entitySet = null;
            var model = CreatePersonModel(out entitySet);

            foreach (string contentType in new string[] { "application/json" })
            {
                var messageWriter = CreateODataMessageWriter(model, contentType);
                var odataWriter = messageWriter.CreateODataResourceWriter();
                Action test = () =>
                {
                    var entry = new ODataResource() { Properties = new List<ODataProperty>(new ODataProperty[] { new ODataProperty() { Name = "Id", Value = 1 } }) };
                    entry.TypeName = "DefaultNamespace.FakeType";
                    odataWriter.WriteStart(entry);
                    odataWriter.WriteEnd();
                };

                test.Throws<ODataException>(Strings.ValidationUtils_UnrecognizedTypeName("DefaultNamespace.FakeType"));
            }
        }

        [Fact]
        public void WriteEntryWithoutTypeNameWithMetadataOnWriterShouldWriteOnJsonLightFailOnOtherFormats()
        {
            IEdmEntitySet entitySet = null;
            var model = CreatePersonModel(out entitySet);

            foreach (string contentType in new string[] { "application/json" })
            {
                var messageWriter = CreateODataMessageWriter(model, contentType);
                var odataWriter = messageWriter.CreateODataResourceWriter(null, entitySet.EntityType());
                Action test = () =>
                {
                    var entry = new ODataResource();
                    odataWriter.WriteStart(entry);
                    odataWriter.WriteEnd();
                };

                test.DoesNotThrow();
            }
        }

        [Fact]
        public void WriteEntryWithNoTypeNameAndNoWriterMetadataShouldFailWithMissingTypeInformation()
        {
            IEdmEntitySet entitySet = null;
            var model = CreatePersonModel(out entitySet);

            foreach (string contentType in new string[] { "application/json" })
            {
                var messageWriter = CreateODataMessageWriter(model, contentType);
                var odataWriter = messageWriter.CreateODataResourceWriter();
                Action test = () =>
                {
                    var entry = new ODataResource() { Properties = new List<ODataProperty>(new ODataProperty[] { new ODataProperty() { Name = "Id", Value = 1 } }) };
                    odataWriter.WriteStart(entry);
                    odataWriter.WriteEnd();
                };

                test.Throws<ODataException>(Strings.WriterValidationUtils_MissingTypeNameWithMetadata);
            }
        }

        [Fact]
        public void WriteEntryOnODataFormatsWithoutSpecifingModelOrUsingEdmCoreModel()
        {
            foreach (IEdmModel model in new IEdmModel[] { null, EdmCoreModel.Instance })
            {
                foreach (string contentType in new string[] { "application/json" })
                {
                    string currentContentType = contentType;
                    var currentModel = model;
                    Action test = () =>
                    {
                        var messageWriter = CreateODataMessageWriter(currentModel, currentContentType);
                        var odataWriter = messageWriter.CreateODataResourceWriter();
                        var entry = new ODataResource() { Properties = new List<ODataProperty>(new ODataProperty[] { new ODataProperty() { Name = "Id", Value = 1 } }) };
                        odataWriter.WriteStart(entry);
                        odataWriter.WriteEnd();
                    };

                    test.DoesNotThrow();
                }
            }
        }

        [Fact]
        public void WriteEntryWithCollectionOfTypeDefinitionShouldWork()
        {
            IEdmEntitySet entitySet = null;
            var model = CreatePersonModelWithTypeDefinition(out entitySet);

            foreach (string contentType in new string[] { "application/json" })
            {
                var messageWriter = CreateODataMessageWriter(model, contentType);
                var odataWriter = messageWriter.CreateODataResourceWriter(null, entitySet.EntityType());
                Action test = () =>
                {
                    var entry = new ODataResource()
                    {
                        Properties = new[]
                        {
                            new ODataProperty()
                            {
                                Name = "Id",
                                Value = 1
                            },
                            new ODataProperty()
                            {
                                Name = "Numbers",
                                Value = new ODataCollectionValue() {Items = new object[] { (UInt32)1, UInt32.MaxValue, UInt32.MinValue } }
                            }
                        }
                    };
                    odataWriter.WriteStart(entry);
                    odataWriter.WriteEnd();
                };

                test.DoesNotThrow();
            }
        }

        private static EdmModel CreatePersonModel(out IEdmEntitySet entitySet)
        {
            var model = new EdmModel();
            var container = new EdmEntityContainer("DefaultNamespace", "Container");
            var personType = new EdmEntityType("DefaultNamespace", "Person");
            var keyProp = personType.AddStructuralProperty("Id", EdmPrimitiveTypeKind.Int32);
            personType.AddKeys(keyProp);
            entitySet = container.AddEntitySet("People", personType);
            model.AddElement(container);
            model.AddElement(personType);
            return model;
        }

        private static EdmModel CreatePersonModelWithTypeDefinition(out IEdmEntitySet entitySet)
        {
            var model = new EdmModel();
            var uint32 = model.GetUInt32("DefaultNamespace", false);
            var container = new EdmEntityContainer("DefaultNamespace", "Container");
            var personType = new EdmEntityType("DefaultNamespace", "Person");
            var keyProp = personType.AddStructuralProperty("Id", EdmPrimitiveTypeKind.Int32);
            personType.AddStructuralProperty("Numbers", new EdmCollectionTypeReference(new EdmCollectionType(uint32)));
            personType.AddKeys(keyProp);
            entitySet = container.AddEntitySet("People", personType);
            model.AddElement(container);
            model.AddElement(personType);
            return model;
        }

        private static ODataMessageWriter CreateODataMessageWriter(IEdmModel model, string contentType)
        {
            var settings = new ODataMessageWriterSettings { EnableMessageStreamDisposal = false, Version = ODataVersion.V4 };
            settings.SetServiceDocumentUri(new Uri("http://example.com"));

            var message = new InMemoryMessage();
            message.Stream = new MemoryStream();
            message.SetHeader("Accept", contentType);
            message.SetHeader("Content-Type", contentType);
            message.SetHeader("OData-Version", "4.0");
            if (model == null)
            {
                return new ODataMessageWriter((IODataRequestMessage)message, settings);
            }
            else
            {
                return new ODataMessageWriter((IODataRequestMessage)message, settings, model);
            }
        }
    }
}
