﻿//---------------------------------------------------------------------
// <copyright file="SearchLexerTests.cs" company="Microsoft">
//      Copyright (C) Microsoft Corporation. All rights reserved. See License.txt in the project root for license information.
// </copyright>
//---------------------------------------------------------------------

using System;
using Microsoft.OData.UriParser;
using Xunit;

namespace Microsoft.OData.Tests.UriParser
{
    public class SearchLexerTests
    {
        [Fact]
        public void BasicTest()
        {
            ExpressionLexer lexer = new SearchLexer("rd");
            ValidateStringLiteralToken(lexer, "rd");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void SpaceTest()
        {
            ExpressionLexer lexer = new SearchLexer("st    dn rd");
            ValidateStringLiteralToken(lexer, "st");
            ValidateStringLiteralToken(lexer, "dn");
            ValidateStringLiteralToken(lexer, "rd");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void ParenTest()
        {
            ExpressionLexer lexer = new SearchLexer("st    (dn   rd)");
            ValidateStringLiteralToken(lexer, "st");
            ValidateTokenKind(lexer, ExpressionTokenKind.OpenParen);
            ValidateStringLiteralToken(lexer, "dn");
            ValidateStringLiteralToken(lexer, "rd");
            ValidateTokenKind(lexer, ExpressionTokenKind.CloseParen);
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void QuoteTest()
        {
            ExpressionLexer lexer = new SearchLexer("a \"AND bc OR \" def");
            ValidateStringLiteralToken(lexer, "a");
            ValidateStringLiteralToken(lexer, "AND bc OR ");
            ValidateStringLiteralToken(lexer, "def");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void QuoteAndEscapeTest()
        {
            ExpressionLexer lexer = new SearchLexer("a \"AND \\\"bc\\\\ OR \" def");
            ValidateStringLiteralToken(lexer, "a");
            ValidateStringLiteralToken(lexer, "AND \"bc\\ OR ");
            ValidateStringLiteralToken(lexer, "def");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void ContinuousQuoteTest()
        {
            ExpressionLexer lexer = new SearchLexer("a \"AND \\\"bc\\\\ OR \"\"VV\" def");
            ValidateStringLiteralToken(lexer, "a");
            ValidateStringLiteralToken(lexer, "AND \"bc\\ OR ");
            ValidateStringLiteralToken(lexer, "VV");
            ValidateStringLiteralToken(lexer, "def");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void TermTest()
        {
            ExpressionLexer lexer = new SearchLexer("a AND bc OR  def");
            ValidateStringLiteralToken(lexer, "a");
            ValidateIdentifierToken(lexer, "AND");
            ValidateStringLiteralToken(lexer, "bc");
            ValidateIdentifierToken(lexer, "OR");
            ValidateStringLiteralToken(lexer, "def");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void InvalidCharWithOutQuoteTest()
        {
            ExpressionLexer lexer = new SearchLexer("\"b\\\"cd a3\"");
            ValidateStringLiteralToken(lexer, "b\"cd a3");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
            lexer = new SearchLexer("\"bcd za3\"");
            ValidateStringLiteralToken(lexer, "bcd za3");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);

            Action action = () => new SearchLexer("b\\\"cd a3").NextToken();
            action.Throws<ODataException>(Strings.ExpressionLexer_InvalidCharacter("\\", 1, "b\\\"cd a3"));
            action = () => new SearchLexer("bcd za3").NextToken();
            action.Throws<ODataException>(Strings.ExpressionLexer_InvalidCharacter("3", 6, "bcd za3"));
            action = () => new SearchLexer("\" za\"\\").NextToken();
            action.Throws<ODataException>(Strings.ExpressionLexer_InvalidCharacter("\\", 5, "\" za\"\\"));
        }

        [Fact]
        public void InvalidEscapeTest()
        {
            Action action = () => new SearchLexer("\"t a\\A\"").NextToken();
            action.Throws<ODataException>(Strings.ExpressionLexer_InvalidEscapeSequence("A", 5, "\"t a\\A\""));
            action = () => new SearchLexer("\"a\\A t\"");
            action.Throws<ODataException>(Strings.ExpressionLexer_InvalidEscapeSequence("A", 3, "\"a\\A t\""));
            action = () => new SearchLexer("\"\\Aa t\"");
            action.Throws<ODataException>(Strings.ExpressionLexer_InvalidEscapeSequence("A", 2, "\"\\Aa t\""));
        }

        [Fact]
        public void EmptyInputTest()
        {
            ExpressionLexer lexer = new SearchLexer("");
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        [Fact]
        public void EmptyPhraseInputTest()
        {
            ExpressionLexer lexer = new SearchLexer("A \"\"");
            Action action = () => lexer.NextToken();
            action.Throws<ODataException>(Strings.ExpressionToken_IdentifierExpected((2)));
            action = () => new SearchLexer("\"\" A");
            action.Throws<ODataException>(Strings.ExpressionToken_IdentifierExpected((0)));
        }

        [Fact]
        public void UnmatchedParenTest()
        {
            ExpressionLexer lexer = new SearchLexer("kit )");
            ValidateStringLiteralToken(lexer, "kit");
            ValidateTokenKind(lexer, ExpressionTokenKind.CloseParen);
            ValidateTokenKind(lexer, ExpressionTokenKind.End);
        }

        private static void ValidateStringLiteralToken(ExpressionLexer lexer, string text)
        {
            Assert.Equal(ExpressionTokenKind.StringLiteral, lexer.CurrentToken.Kind);
            Assert.Equal(text, lexer.CurrentToken.Text);
            lexer.NextToken();
        }

        private static void ValidateIdentifierToken(ExpressionLexer lexer, string text)
        {
            Assert.Equal(ExpressionTokenKind.Identifier, lexer.CurrentToken.Kind);
            Assert.Equal(text, lexer.CurrentToken.Text);
            lexer.NextToken();
        }

        private static void ValidateTokenKind(ExpressionLexer lexer, ExpressionTokenKind kind)
        {
            Assert.Equal(kind, lexer.CurrentToken.Kind);
            lexer.NextToken();
        }
    }
}
